import numpy
import math

from amuse.units import nbody_system, units, constants, quantities
from amuse.datamodel import Particles, Particle
from amuse.ic.plummer import new_plummer_model
from amuse.couple.bridge import Bridge, CalculateFieldForCodesUsingReinitialize

from amuse.test.amusetest import TestWithMPI
from amuse.community.fastkick.interface import FastKickInterface, FastKick


class TestFastKickInterface(TestWithMPI):

    number_of_workers = 2
    mode = "cpu"

    def test1(self):
        instance = FastKickInterface(mode=self.mode, number_of_workers=self.number_of_workers)
        instance.initialize_code()
        id1, error1 = instance.new_particle(mass=11.0, x=0.0, y=0.0, z=0.0)
        id2, error2 = instance.new_particle(mass=21.0, x=10.0, y=0.0, z=0.0)
        self.assertEqual(0, id1)
        self.assertEqual(0, error1)
        self.assertEqual(1, id2)
        self.assertEqual(0, error2)
        self.assertEqual(0, instance.commit_particles())
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test2(self):
        fastkick = FastKickInterface(mode=self.mode, number_of_workers=self.number_of_workers)
        self.assertEqual(0, fastkick.set_eps2(0.101))
        self.assertEqual([0.101, 0], list(fastkick.get_eps2().values()))
        self.assertEqual(0, fastkick.set_eps2(0.2))
        self.assertEqual([0.2, 0], list(fastkick.get_eps2().values()))
        fastkick.cleanup_code()
        fastkick.stop()

    def test3(self):
        fastkick = FastKickInterface(mode=self.mode, number_of_workers=self.number_of_workers)
        fastkick.initialize_code()
        fastkick.new_particle([10, 10], [-1, 1], [0, 0], [0, 0])
        self.assertEqual(0, fastkick.commit_particles())
        self.assertEqual([-20.0, 0], list(fastkick.get_potential_at_point(0, 0, 0, 0).values()))
        self.assertAlmostEqual(-10.0*math.sqrt(2.0), list(fastkick.get_potential_at_point(1.0, 0, 0, 0).values())[0], 4)
        fastkick.cleanup_code()
        fastkick.stop()

    def test4(self):
        numpy.random.seed(12345)
        plummer = new_plummer_model(100)
        fastkick = FastKickInterface(mode=self.mode, number_of_workers=self.number_of_workers)
        fastkick.initialize_code()
        fastkick.new_particle([1]*100, plummer.x.number, plummer.y.number, plummer.z.number)
        self.assertEqual(0, fastkick.commit_particles())
        points = new_plummer_model(73)
        potential1, error = fastkick.get_potential_at_point([0]*73,
            points.x.number, points.y.number, points.z.number)
        ax1, ay1, az1, error = fastkick.get_gravity_at_point([0]*73,
            points.x.number, points.y.number, points.z.number)
        fastkick.cleanup_code()

        fastkick.initialize_code()
        self.assertEqual(0, fastkick.commit_particles())
        potential0, error = fastkick.get_potential_at_point([0]*73,
            points.x.number, points.y.number, points.z.number)
        ax0, ay0, az0, error = fastkick.get_gravity_at_point([0]*73,
            points.x.number, points.y.number, points.z.number)
        self.assertAlmostEqual(potential0, 0)
        self.assertAlmostEqual(ax0, 0)
        self.assertAlmostEqual(ay0, 0)
        self.assertAlmostEqual(az0, 0)
        fastkick.cleanup_code()

        fastkick.initialize_code()
        for p in plummer:
            fastkick.new_particle(1, p.x.number, p.y.number, p.z.number)
        self.assertEqual(0, fastkick.commit_particles())
        potential2, error = fastkick.get_potential_at_point([0]*73,
            points.x.number, points.y.number, points.z.number)
        ax2, ay2, az2, error = fastkick.get_gravity_at_point([0]*73,
            points.x.number, points.y.number, points.z.number)
        self.assertAlmostEqual(potential1, potential2, 4)
        self.assertAlmostEqual(ax1, ax2, 4)
        self.assertAlmostEqual(ay1, ay2, 4)
        self.assertAlmostEqual(az1, az2, 4)
        fastkick.cleanup_code()
        fastkick.stop()


class TestFastKick(TestWithMPI):

    mode = "cpu"

    def new_fastkick_instance(self, convert_nbody=None, number_of_workers=2, **kwargs):
        return FastKick(convert_nbody, mode=self.mode, number_of_workers=number_of_workers, **kwargs)

    def test1(self):
        print("Testing FastKick (SI)")
        sun = Particle()
        sun.mass = 1.0 | units.MSun
        sun.position = [0, 0, 0] | units.m
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        fastkick = self.new_fastkick_instance(convert_nbody)
        fastkick.particles.add_particle(sun)
        ax, ay, az = fastkick.get_gravity_at_point(0 | units.AU, 1 | units.AU, 0 | units.AU, 0 | units.AU)
        fastkick.stop()
        self.assertAlmostRelativeEqual(ax, -4*math.pi**2 | units.AU / units.yr**2, 3)
        self.assertAlmostRelativeEqual(ay, 0 | units.m * units.s**-2, 6)
        self.assertAlmostRelativeEqual(az, 0 | units.m * units.s**-2, 6)

    def test2(self):
        print("Testing FastKick reset")
        particles = new_plummer_model(50)
        instance = self.new_fastkick_instance()
        instance.parameters.epsilon_squared = 0.12345 | nbody_system.length**2
        instance.particles.add_particles(particles)
        self.assertEqual(len(instance.particles), 50)

        instance.reset()

        self.assertAlmostRelativeEquals(instance.parameters.epsilon_squared, 0.12345 | nbody_system.length**2)
        self.assertEqual(len(instance.particles), 0)
        instance.particles.add_particles(particles)
        self.assertEqual(len(instance.particles), 50)

    def test3(self):
        print("Testing FastKick parameters")
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.yr, 1.0 | units.AU)
        instance = self.new_fastkick_instance(convert_nbody)

        self.assertAlmostEqual(0.0 | units.AU**2, instance.parameters.epsilon_squared, in_units=units.AU**2)
        for x in [0.01, 0.1, 0.2]:
            instance.parameters.epsilon_squared = x | units.AU**2
            self.assertAlmostEqual(x | units.AU**2, instance.parameters.epsilon_squared, in_units=units.AU**2)

        instance.stop()

    def test4(self):
        numpy.random.seed(12345)
        plummer = new_plummer_model(100)
        fastkick = self.new_fastkick_instance()
        fastkick.initialize_code()
        fastkick.particles.add_particles(plummer)
        points = new_plummer_model(73)
        potential1 = fastkick.get_potential_at_point(0*points.x, points.x, points.y, points.z)
        ax1, ay1, az1 = fastkick.get_gravity_at_point(0*points.x, points.x, points.y, points.z)
        fastkick.reset()

        potential0 = fastkick.get_potential_at_point(0*points.x, points.x, points.y, points.z)
        ax0, ay0, az0 = fastkick.get_gravity_at_point(0*points.x, points.x, points.y, points.z)
        self.assertAlmostEqual(potential0, 0 | nbody_system.potential)
        self.assertAlmostEqual(ax0, 0 | nbody_system.acceleration)
        self.assertAlmostEqual(ay0, 0 | nbody_system.acceleration)
        self.assertAlmostEqual(az0, 0 | nbody_system.acceleration)
        fastkick.reset()

        for p in plummer:
            fastkick.particles.add_particle(p)
        potential2 = fastkick.get_potential_at_point(0*points.x, points.x, points.y, points.z)
        ax2, ay2, az2 = fastkick.get_gravity_at_point(0*points.x, points.x, points.y, points.z)
        self.assertAlmostEqual(potential1, potential2, 5)
        self.assertAlmostEqual(ax1, ax2, 5)
        self.assertAlmostEqual(ay1, ay2, 5)
        self.assertAlmostEqual(az1, az2, 5)
        fastkick.cleanup_code()
        fastkick.stop()

    def test5(self):
        print("Testing FastKick states")
        plummer = new_plummer_model(100)

        print("First do everything manually:")
        instance = self.new_fastkick_instance()
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.initialize_code()
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.commit_parameters()
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        instance.particles.add_particles(plummer)
        instance.commit_particles()
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.cleanup_code()
        self.assertEqual(instance.get_name_of_current_state(), 'END')
        instance.stop()

        print("commit_parameters(), (re)commit_particles(), and cleanup_code() should be called "
            "automatically before new_xx_particle(), get_xx(), and stop():")
        instance = self.new_fastkick_instance()
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        eps = instance.parameters.epsilon_squared
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.particles.add_particles(plummer[:-1])
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        acc = instance.get_gravity_at_point(0*plummer[-1].x, plummer[-1].x, plummer[-1].y, plummer[-1].z)
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.particles.add_particle(plummer[-1])
        self.assertEqual(instance.get_name_of_current_state(), 'UPDATE')
        acc = instance.get_gravity_at_point(0*plummer[-1].x, plummer[-1].x, plummer[-1].y, plummer[-1].z)
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.stop()
        self.assertEqual(instance.get_name_of_current_state(), 'STOPPED')

    def test6(self):
        plummer = new_plummer_model(100)
        points = new_plummer_model(73)
        instance = self.new_fastkick_instance(number_of_workers=1)
        instance.particles.add_particles(plummer)
        potential1 = instance.get_potential_at_point(0*points.x, points.x, points.y, points.z)
        ax1, ay1, az1 = instance.get_gravity_at_point(0*points.x, points.x, points.y, points.z)
        instance.stop()

        expected_accuracy = 13 if self.mode == "cpu" else 5
        number_of_workers_range = [2, 3, 4, 5] if self.mode == "cpu" else [2]
        for n in number_of_workers_range:
            instance = self.new_fastkick_instance(number_of_workers=n)
            instance.particles.add_particles(plummer)
            potential = instance.get_potential_at_point(0*points.x, points.x, points.y, points.z)
            ax, ay, az = instance.get_gravity_at_point(0*points.x, points.x, points.y, points.z)
            instance.stop()

            self.assertAlmostEqual(potential, potential1, expected_accuracy)
            self.assertAlmostEqual(ax, ax1, expected_accuracy)
            self.assertAlmostEqual(ay, ay1, expected_accuracy)
            self.assertAlmostEqual(az, az1, expected_accuracy)

    def test7(self):
        instance = self.new_fastkick_instance()
        instance.parameters.epsilon_squared = 0.00001 | nbody_system.length**2

        particles = Particles(2)
        particles.mass = [1.0, 1.0] | nbody_system.mass
        particles.position = [[0.0, 0.0, 0.0], [2.0, 0.0, 0.0]] | nbody_system.length
        instance.particles.add_particles(particles)

        zero = 0.0 | nbody_system.length
        ax, ay, az = instance.get_gravity_at_point(zero, 1.0 | nbody_system.length, zero, zero)
        self.assertAlmostEqual(ax, 0.0 | nbody_system.acceleration, 6)
        self.assertAlmostEqual(ay, 0.0 | nbody_system.acceleration, 6)
        self.assertAlmostEqual(az, 0.0 | nbody_system.acceleration, 6)

        for x in (0.25, 0.5, 0.75):
            x0 = x | nbody_system.length
            x1 = (2.0 - x) | nbody_system.length
            potential0 = instance.get_potential_at_point(zero, x0, zero, zero)
            potential1 = instance.get_potential_at_point(zero, x1, zero, zero)
            ax0, ay0, az0 = instance.get_gravity_at_point(zero, x0, zero, zero)
            ax1, ay1, az1 = instance.get_gravity_at_point(zero, x1, zero, zero)

            self.assertAlmostEqual(ay0, 0.0 | nbody_system.acceleration, 6)
            self.assertAlmostEqual(az0, 0.0 | nbody_system.acceleration, 6)
            self.assertAlmostEqual(ay1, 0.0 | nbody_system.acceleration, 6)
            self.assertAlmostEqual(az1, 0.0 | nbody_system.acceleration, 6)

            self.assertAlmostEqual(ax0, -ax1, 5)
            ax = (-1.0 / (x0**2) + 1.0 / (x1**2)) * (1.0 | nbody_system.length**3 / nbody_system.time**2)
            self.assertAlmostEqual(ax, ax0, 2)
            self.assertAlmostEqual(potential0, potential1, 5)
        instance.stop()

    def test8(self):
        print("Testing FastKick for Bridge: potential energy")
        numpy.random.seed(12345)
        stars = new_plummer_model(100)
        cluster = new_gravity_code(stars)

        first_half = stars.select_array(lambda x: (x > 0 | nbody_system.length), ['x'])
        second_half = stars - first_half
        cluster1 = new_gravity_code(first_half)
        cluster2 = new_gravity_code(second_half)
        kick_from_cluster1 = CalculateFieldForCodesUsingReinitialize(self.new_fastkick_instance(), (cluster1,))
        kick_from_cluster2 = CalculateFieldForCodesUsingReinitialize(self.new_fastkick_instance(), (cluster2,))
        bridgesys = Bridge()
        bridgesys.add_system(cluster1, (kick_from_cluster2,))
        bridgesys.add_system(cluster2, (kick_from_cluster1,))

        self.assertAlmostRelativeEqual(cluster.potential_energy, bridgesys.potential_energy, 7)
        self.assertAlmostRelativeEqual(cluster.kinetic_energy, bridgesys.kinetic_energy)
        kick_from_cluster1.code.stop()
        kick_from_cluster2.code.stop()

    def test9(self):
        print("Testing FastKick for Bridge: evolving a binary")
        particles = Particles(2)
        particles.mass = [3.0, 1.0] | units.MSun
        particles.position = [0, 0, 0] | units.AU
        particles.velocity = [0, 0, 0] | units.km / units.s
        particles[1].x = 2.0 | units.AU
        particles[1].vy = (constants.G * (4.0 | units.MSun) / (2.0 | units.AU)).sqrt()
        particles.move_to_center()

        primary_sys = new_gravity_code(particles[:1])
        secondary_sys = new_gravity_code(particles[1:])

        primary = primary_sys.particles[0]
        P = 2 * math.pi * primary.x / primary.vy

        converter = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        kick_from_primary = CalculateFieldForCodesUsingReinitialize(self.new_fastkick_instance(converter), (primary_sys,))
        kick_from_secondary = CalculateFieldForCodesUsingReinitialize(self.new_fastkick_instance(converter), (secondary_sys,))

        bridgesys = Bridge(timestep=P / 64.0)
        bridgesys.add_system(primary_sys, (kick_from_secondary,))
        bridgesys.add_system(secondary_sys, (kick_from_primary,))

        position_at_start = primary.position.x
        bridgesys.evolve_model(P / 4.0)
        self.assertAlmostRelativeEqual(position_at_start, primary.position.y, 2)

        bridgesys.evolve_model(P / 2.0)
        self.assertAlmostRelativeEqual(position_at_start, -primary.position.x, 2)

        bridgesys.evolve_model(P)
        kick_from_primary.code.stop()
        kick_from_secondary.code.stop()
        self.assertAlmostRelativeEqual(position_at_start, primary.position.x, 2)

    def test10(self):
        number_of_sources = 10
        number_of_points = 1000000
        mass, length, G = nbody_system.mass, nbody_system.length, nbody_system.G
        sources = Particles(mass=numpy.ones(number_of_sources) | mass, x=0 | length, y=0 | length, z=0 | length)
        points = Particles(x=0 | length, y=0 | length, z=numpy.arange(number_of_points)+1.0 | length)

        instance = self.new_fastkick_instance()
        instance.particles.add_particles(sources)
        potential = instance.get_potential_at_point(0*points.x, points.x, points.y, points.z)
        ax, ay, az = instance.get_gravity_at_point(0*points.x, points.x, points.y, points.z)
        self.assertAlmostEqual(ax, G * (0 | mass/length**2), 5)
        self.assertAlmostEqual(ay, G * (0 | mass/length**2), 5)
        self.assertAlmostRelativeEqual(az, -G*sources.total_mass()/points.z**2, 3)
        self.assertAlmostRelativeEqual(potential, -G*sources.total_mass()/points.z, 3)
        instance.stop()

    def test11(self):
        print("Test that a source is not included when calculating gravity on itself.")
        number_of_sources = 100
        mass, length, G = nbody_system.mass, nbody_system.length, nbody_system.G
        sources = Particles(mass=numpy.ones(number_of_sources) | mass, x=0 | length, y=0 | length, z=0 | length)
        point = Particle(x=0 | length, y=0 | length, z=1.0 | length)

        instance = self.new_fastkick_instance()
        instance.particles.add_particles(sources)
        potential = instance.get_potential_at_point(0 | length, point.x, point.y, point.z)
        ax, ay, az = instance.get_gravity_at_point(0 | length, point.x, point.y, point.z)
        self.assertAlmostEqual(ax, G * (0 | mass/length**2), 5)
        self.assertAlmostEqual(ay, G * (0 | mass/length**2), 5)
        self.assertAlmostRelativeEqual(az, -G*sources.total_mass()/point.z**2, 3)
        self.assertAlmostRelativeEqual(potential, -G*sources.total_mass()/point.z, 3)

        point.mass = 1e6 | mass
        instance.particles.add_particle(point)
        potential = instance.get_potential_at_point(0 | length, point.x, point.y, point.z)
        ax, ay, az = instance.get_gravity_at_point(0 | length, point.x, point.y, point.z)
        self.assertAlmostEqual(ax, G * (0 | mass/length**2), 5)
        self.assertAlmostEqual(ay, G * (0 | mass/length**2), 5)
        self.assertAlmostRelativeEqual(az, -G*sources.total_mass()/point.z**2, 3)
        self.assertAlmostRelativeEqual(potential, -G*sources.total_mass()/point.z, 3)
        instance.stop()

    def test12(self):
        print("Test FastKick potential energy")
        number_of_sources = 1000
        numpy.random.seed(12345)
        plummer = new_plummer_model(number_of_sources)

        fastkick = self.new_fastkick_instance()
        fastkick.particles.add_particles(plummer)
        self.assertAlmostRelativeEqual(fastkick.potential_energy, plummer.potential_energy(G=nbody_system.G), 6)
        fastkick.cleanup_code()
        fastkick.stop()

    def test13(self):
        number_of_sources = 10
        mass, length, G = nbody_system.mass, nbody_system.length, nbody_system.G
        sources = Particles(
                mass=numpy.ones(number_of_sources) | mass,
                x=0 | length,
                y=0 | length,
                z=0 | length)

        instance = self.new_fastkick_instance()
        instance.particles.add_particles(sources)
        self.assertAlmostRelativeEqual(instance.particles.mass, 1 | mass)
        instance.particles.mass = 2 | mass
        self.assertAlmostRelativeEqual(instance.particles.mass, 2 | mass)


class TestFastKickGPU(TestFastKick):

    mode = "gpu"

    def new_fastkick_instance(self, convert_nbody=None, number_of_workers=1, **kwargs):
        return self.new_instance_of_an_optional_code(FastKick, convert_nbody,
            mode=self.mode, number_of_workers=number_of_workers, **kwargs)


class GravityCodeForTesting(object):

    def __init__(self):
        self.particles = Particles()
        self.model_time = quantities.zero

    def evolve_model(self, t_end):
        self.particles.position += self.particles.velocity * (t_end - self.model_time)
        self.model_time = t_end

    @property
    def potential_energy(self):
        G = nbody_system.G if self.particles.x.unit == nbody_system.length else constants.G
        return self.particles.potential_energy(G=G)

    @property
    def kinetic_energy(self):
        return self.particles.kinetic_energy()


def new_gravity_code(particles):
    instance = GravityCodeForTesting()
    instance.particles.add_particles(particles)
    return instance



from amuse.test import amusetest

from amuse.support.exceptions import AmuseException
from amuse.units import units
from amuse.units import constants
from amuse.units import nbody_system
from amuse.support.interface import InCodeComponentImplementation
import numpy
import time
import sys

from amuse.datamodel import trees
from amuse.datamodel import Particles


class TestBinaryTree(amusetest.TestCase):

    def test1(self):
        particles = Particles(10)
        particles.mass = range(10) | units.kg
        particles[0].child1 = particles[1]
        particles[0].child2 = particles[2]

        self.assertEqual(particles[0].mass, 0 | units.kg)
        self.assertEqual(particles[0].child1.mass, 1 | units.kg)
        self.assertEqual(particles[1].child1, None)

        children1 = particles.child1.as_set().compressed()
        children2 = particles.child2.as_set().compressed()
        children = children1 + children2
        roots = particles - children

        self.assertEqual(len(roots), 8)
        self.assertEqual(len(children), 2)

    def test2(self):
        n = 100000
        particles = Particles(n)
        particles.mass = range(n) | units.kg
        particles[n-1].child1 = particles[0]
        particles[n-1].child2 = particles[1]

        self.assertEqual(particles[0].mass, 0 | units.kg)
        self.assertEqual(particles[n-1].child1.mass, 0 | units.kg)
        self.assertEqual(particles[n-1].child2.mass, 1 | units.kg)

        children1 = particles.child1.as_set().compressed()
        children2 = particles.child2.as_set().compressed()
        children = children1 + children2
        roots = particles - children

        self.assertEqual(len(roots), n - 2)
        self.assertEqual(len(children), 2)
        binaries = particles.select_array(lambda x: x != [None], ['child1',])
        self.assertEqual(len(binaries), 1)

    def test3(self):
        n = 10
        particles = Particles(n)
        particles.mass = range(n) | units.kg
        particles[0].child1 = particles[1]
        particles[0].child2 = particles[2]
        particles[1].child1 = particles[3]
        particles[1].child2 = particles[4]

        self.assertEqual(particles[0].child1.child1.mass, 3 | units.kg)

        binaries = particles.select_array(lambda x: x != [None], ["child1",])

        self.assertEqual(len(binaries), 2)

        binaries_children1 = binaries.child1.as_set().compressed().select_array(lambda x: x != [None], ["child1",])
        binaries_children2 = binaries.child2.as_set().compressed().select_array(lambda x: x != [None], ["child1",])
        binaries_roots = binaries - (binaries_children1 + binaries_children2)

        self.assertEqual(len(binaries_roots), 1)

        self.assertEqual(binaries_roots[0].mass, 0 | units.kg)

    def test4(self):
        particles = Particles(10)
        particles.mass = range(10) | units.kg
        particles[0].child1 = particles[1]
        particles[0].child2 = particles[2]
        particles[1].child1 = particles[3]
        particles[1].child2 = particles[4]

        x = trees.BinaryTreesOnAParticleSet(particles, "child1", "child2")
        roots = list(x.iter_roots())

        self.assertEqual(len(roots), 1)
        x = [x.mass.value_in(units.kg) for x in roots[0].iter_descendants()]
        self.assertEqual(x, [1, 2, 3, 4])
        self.assertEqual(roots[0].get_descendants_subset().mass, [1, 2, 3, 4] | units.kg)

    def test5(self):
        particles = Particles(10)
        particles.mass = range(10) | units.kg
        particles[0].child1 = particles[1]
        particles[0].child2 = particles[2]
        particles[1].child1 = particles[3]
        particles[1].child2 = particles[4]

        x = trees.BinaryTreesOnAParticleSet(particles, "child1", "child2")
        roots = list(x.iter_roots())

        self.assertEqual(len(roots), 1)
        y = [(event, x.mass.value_in(units.kg)) for event, x in roots[0].iter_events()]
        self.assertEqual(y,
            [
                ('start', 0.0),
                    ('start', 1.0),
                        ('start', 3.0),
                        ('end', 3.0),
                        ('start', 4.0),
                        ('end', 4.0),
                    ('end', 1.0),
                    ('start', 2.0),
                    ('end', 2.0),
                ('end', 0.0)
            ]
        )

    def test6(self):

        particles = Particles(10)
        particles.mass = range(10) | units.kg
        particles[0].child1 = particles[1]
        particles[0].child2 = particles[2]
        particles[1].child1 = particles[3]
        particles[1].child2 = particles[4]

        x = trees.BinaryTreesOnAParticleSet(particles, "child1", "child2")
        roots = list(x.iter_roots())

        self.assertEqual(len(roots), 1)
        y = [(event, x.mass.value_in(units.kg)) for event, x in roots[0].iter_levels()]
        self.assertEqual(y,
            [
                (0, 0.0),
                    (1, 1.0),
                        (2, 3.0),
                        (2, 4.0),
                    (1, 2.0),
            ]
        )

    def test7(self):

        particles = Particles(10)
        particles.mass = range(10) | units.kg
        particles[0].child1 = particles[1]
        particles[0].child2 = particles[2]
        particles[1].child1 = particles[3]
        particles[1].child2 = particles[4]

        x = trees.BinaryTreesOnAParticleSet(particles, "child1", "child2")
        roots = list(x.iter_roots())

        self.assertEqual(len(roots), 1)
        binary = roots[0]
        output = ''
        for level, particle in binary.iter_levels():
            output += '..' * level
            output += str(particle.mass.value_in(units.kg))
            output += '\n'

        self.assertEqual(output, """0.0
..1.0
....3.0
....4.0
..2.0
""")

    def test8(self):
        particles = Particles(10)
        particles.mass = range(10) | units.kg
        particles[0].child1 = particles[1]
        particles[0].child2 = particles[2]
        particles[1].child1 = particles[3]
        particles[1].child2 = particles[4]

        x = trees.BinaryTreesOnAParticleSet(particles, "child1", "child2")
        self.assertEqual(len(list(x.iter_roots())), 1)
        self.assertEqual(len(x.particles_not_in_a_multiple()), 5)


class TestChildTree(amusetest.TestCase):

    def test1(self):
        particles = Particles(10)
        particles.mass = range(10) | units.kg
        particles[0].child1 = particles[1]
        particles[0].child2 = particles[2]

        self.assertEqual(particles[0].mass, 0 | units.kg)
        self.assertEqual(particles[0].child1.mass, 1 | units.kg)
        self.assertEqual(particles[1].child1, None)

        tree = particles.as_binary_tree()
        self.assertFalse(tree.is_leaf())
        self.assertEqual(len(list(tree.iter_children())), 8)
        self.assertEqual(len(list(tree.iter_branches())), 1)
        self.assertEqual(len(list(tree.iter_leafs())), 7)
        branches = list(tree.iter_branches())
        self.assertEqual(len(list(branches[0].iter_children())), 2)
        self.assertEqual(len(list(branches[0].iter_branches())), 0)
        self.assertEqual(len(list(branches[0].iter_leafs())), 2)

    def test2(self):
        particles = Particles(10)
        particles.mass = range(10) | units.kg
        particles[0].child1 = particles[1]
        particles[0].child2 = particles[2]

        self.assertEqual(particles[0].mass, 0 | units.kg)
        self.assertEqual(particles[0].child1.mass, 1 | units.kg)
        self.assertEqual(particles[1].child1, None)

        tree = particles.as_binary_tree()
        self.assertFalse(tree.is_leaf())
        self.assertEqual(len(list(tree.iter_descendant_leafs())), 9)
        self.assertEqual(len(list(tree.iter_descendant_branches())), 1)
        branches = list(tree.iter_branches())
        self.assertEqual(len(list(branches[0].iter_descendant_leafs())), 2)
        self.assertEqual(len(list(branches[0].iter_descendant_branches())), 0)

    def test3(self):
        particles = Particles(10)
        particles.mass = range(10) | units.kg
        particles[0].child1 = particles[1]
        particles[0].child2 = particles[2]
        particles[1].child1 = particles[3]
        particles[1].child2 = particles[4]
        particles[2].child1 = particles[5]
        particles[2].child2 = particles[6]

        tree = particles.as_binary_tree()
        self.assertFalse(tree.is_leaf())
        self.assertEqual(len(list(tree.iter_children())), 4)
        self.assertEqual(len(list(tree.iter_branches())), 1)
        self.assertEqual(len(list(tree.iter_leafs())), 3)
        self.assertEqual(len(list(tree.iter_descendant_leafs())), 7)
        self.assertEqual(len(list(tree.iter_descendant_branches())), 3)
        branches = list(tree.iter_branches())
        self.assertEqual(len(list(branches[0].iter_children())), 2)
        self.assertEqual(len(list(branches[0].iter_branches())), 2)
        self.assertEqual(len(list(branches[0].iter_leafs())), 0)
        self.assertEqual(len(list(branches[0].iter_descendant_leafs())), 4)
        self.assertEqual(len(list(branches[0].iter_descendant_branches())), 2)

    def test4(self):
        particles = Particles(10)
        particles.mass = range(10) | units.kg
        particles[0].child1 = particles[2]
        particles[0].child2 = particles[1]
        particles[1].child1 = particles[4]
        particles[1].child2 = particles[3]
        particles[2].child1 = particles[6]
        particles[2].child2 = particles[5]

        tree = particles.as_binary_tree()
        masses = [] | units.kg
        stack = list(reversed(tree.get_children()))
        while stack:
            x = stack.pop()
            masses.append(x.particle.mass)
            stack.extend(reversed(x.get_children()))

        self.assertEqual(masses, [0, 2, 6, 5, 1, 4, 3, 7, 8, 9] | units.kg)

    def test5(self):

        particles = Particles(10)
        particles.mass = range(10) | units.kg
        particles[0].child1 = particles[1]
        particles[0].child2 = particles[2]
        particles[1].child1 = particles[3]
        particles[1].child2 = particles[4]

        x = particles.as_binary_tree()
        roots = list(x.iter_branches())

        self.assertEqual(len(roots), 1)
        binary = roots[0]
        output = ''
        for level, node in binary.iter_levels():
            output += '..' * level
            output += str(node.particle.mass.value_in(units.kg))
            output += '\n'

        self.assertEqual(output, """0.0
..1.0
....3.0
....4.0
..2.0
""")



from amuse.test import amusetest

from amuse.ext.evrard_test import new_evrard_gas_sphere


from amuse.units import nbody_system as nbody
from amuse.units import units


class TestEvrardModel(amusetest.TestCase):
    def test1(self):
        print("First test: making an Evrard gas sphere model.")
        target_number_of_particles = 1000
        gas_parts = new_evrard_gas_sphere(target_number_of_particles, seed=1234)
        self.assertEqual(len(gas_parts), 1000)

    def test2(self):
        print("Testing properties of an Evrard model.")
        target_number_of_particles = 1000
        gas_parts = new_evrard_gas_sphere(target_number_of_particles, do_scale=True, seed=1234)
        self.assertEqual(len(gas_parts), 1000)
        self.assertAlmostEqual(gas_parts.kinetic_energy(),             0.00 | nbody.energy)
        self.assertAlmostEqual(gas_parts.potential_energy(G=nbody.G), -0.50 | nbody.energy)
        self.assertAlmostEqual(gas_parts.center_of_mass(),          [0, 0, 0] | nbody.length)
        self.assertAlmostEqual(gas_parts.center_of_mass_velocity(), [0, 0, 0] | nbody.speed)
        self.assertAlmostEqual(gas_parts.mass.sum(),                   1.00 | nbody.mass)

    def test3(self):
        print("Testing virial radius of an Evrard model.")
        target_number_of_particles = 100
        gas_parts = new_evrard_gas_sphere(target_number_of_particles, do_scale=True, seed=1234)
        self.assertAlmostEqual(gas_parts.virial_radius(),             1.00 | nbody.length)



from amuse.test import amusetest
from io import StringIO
import textwrap
import os
import numpy


from amuse import io
from amuse.io import vtk
from amuse.units import units
from amuse.units import generic_unit_system
from amuse.units import generic_unit_system
from amuse import datamodel


class VtkStructuredGridTests(amusetest.TestCase):

    def test1(self):
        grid = datamodel. new_regular_grid([2, 3, 4], [1, 1, 1] | generic_unit_system.length)
        grid.rho = grid.x * (0.1 | generic_unit_system.mass / generic_unit_system.length ** 4)
        data_file = StringIO()
        instance = vtk.VtkStructuredGrid("test.vts", data_file, grid)
        instance.store()

        contents = data_file.getvalue()
        self.assertTrue(contents.find('WholeExtent="0 2 0 3 0 4"') > 0)
        self.assertTrue(contents.find('<DataArray type="Float64" NumberOfComponents="3">') > 0)
        self.assertTrue(contents.find('<DataArray type="Float64" NumberOfComponents="1" Name="rho">') > 0)

    def test2(self):
        grid = datamodel. new_regular_grid([4, 5, 6], [1, 1, 1] | generic_unit_system.length)
        grid.mass = generic_unit_system.density(numpy.random.rand(4, 5, 6))
        data_file = StringIO()
        instance = vtk.VtkStructuredGrid("test.vts", data_file, grid)
        instance.store()

        contents = data_file.getvalue()
        self.assertTrue(contents.find('<Piece Extent="0 4 0 5 0 6">') > 0)
        self.assertTrue(contents.find('<CellData>') > 0)
        self.assertTrue(contents.find('<PointData>') > 0)
        self.assertTrue(contents.find('<DataArray type="Float64" NumberOfComponents="3">') > 0)
        self.assertTrue(contents.find('<DataArray type="Float64" NumberOfComponents="1" Name="mass">') > 0)


class VtkUnstructuredGridTests(amusetest.TestCase):

    def test1(self):
        grid = datamodel. new_regular_grid([2, 3, 4], [1, 1, 1] | generic_unit_system.length)
        grid.rho = grid.x * (0.1 | generic_unit_system.mass / generic_unit_system.length ** 4)
        data_file = StringIO()
        instance = vtk.VtkUnstructuredGrid("test.vtu", data_file, grid)
        instance.store()
        contents = data_file.getvalue()

        self.assertTrue(contents.find('<Piece NumberOfPoints="60" NumberOfCells="24"') > 0)
        self.assertTrue(contents.find('DataArray type="Int32" NumberOfComponents="1" Name="connectivity">') > 0)
        self.assertTrue(contents.find('DataArray type="Int32" NumberOfComponents="1" Name="offsets">') > 0)
        self.assertTrue(contents.find('11 11 11') > 0)
        self.assertTrue(contents.find('8 16 24') > 0)
        self.assertTrue(contents.find('DataArray type="Float64" NumberOfComponents="3" Name="points">') > 0)

    def test2(self):
        grid = datamodel. new_regular_grid([2, 2, 2], [1, 1, 1] | generic_unit_system.length)
        grid.rho = grid.x * (0.1 | generic_unit_system.mass / generic_unit_system.length ** 4)
        data_file = StringIO()
        instance = vtk.VtkUnstructuredGrid("test.vtu", data_file, grid)
        instance.store()
        contents = data_file.getvalue()

        self.assertTrue(contents.find('<Piece NumberOfPoints="27" NumberOfCells="8"') > 0)
        self.assertTrue(contents.find('DataArray type="Int32" NumberOfComponents="1" Name="connectivity">') > 0)
        self.assertTrue(contents.find('DataArray type="Int32" NumberOfComponents="1" Name="offsets">') > 0)
        self.assertTrue(contents.find('11 11 11') > 0)
        self.assertTrue(contents.find('56 64') > 0)
        self.assertTrue(contents.find('DataArray type="Float64" NumberOfComponents="3" Name="points">') > 0)
        self.assertTrue(contents.find('<DataArray type="Float64" NumberOfComponents="1" Name="rho">') > 0)

    def test3(self):
        grid1 = datamodel. new_regular_grid([2, 2, 2], [1, 1, 1] | generic_unit_system.length)
        grid1.rho = grid1.x * (0.1 | generic_unit_system.mass / generic_unit_system.length ** 4)
        grid2 = datamodel. new_regular_grid([2, 2, 2], [1, 1, 1] | generic_unit_system.length)
        grid2.position += [1.0, 0.0, 0.0] | generic_unit_system.length
        grid2.rho = grid2.x * (0.1 | generic_unit_system.mass / generic_unit_system.length ** 4)

        data_file = StringIO()
        instance = vtk.VtkUnstructuredGrid("test.vtu", data_file, [grid1, grid2])
        instance.store()
        contents = data_file.getvalue()
        index_of_piece_1 = contents.find('<Piece NumberOfPoints="27" NumberOfCells="8"')
        index_of_piece_2 = contents.find('<Piece NumberOfPoints="27" NumberOfCells="8"', index_of_piece_1 + 1)
        self.assertTrue(index_of_piece_1 > 0)
        self.assertTrue(index_of_piece_2 > 0)
        self.assertTrue(contents.find('DataArray type="Int32" NumberOfComponents="1" Name="connectivity">') > 0)
        self.assertTrue(contents.find('DataArray type="Int32" NumberOfComponents="1" Name="offsets">') > 0)
        self.assertTrue(contents.find('11 11 11') > 0)
        self.assertTrue(contents.find('56 64') > 0)
        self.assertTrue(contents.find('DataArray type="Float64" NumberOfComponents="3" Name="points">') > 0)
        self.assertTrue(contents.find('<DataArray type="Float64" NumberOfComponents="1" Name="rho">') > 0)



from amuse.test import amusetest

from amuse.ext.job_server import RemoteCodeInterface, JobServer

from socket import gethostname
import os

from amuse.ic.plummer import new_plummer_model


def example_parallel_jobs(N, Nhosts=1):

    jobserver = JobServer(hosts=[gethostname()]*Nhosts)

    for i in range(1, N+1):
        jobserver.submit_job(new_plummer_model, (i,))

    result = dict()
    while jobserver.wait():
        job = jobserver.last_finished_job
        result[job.args[0]] = job.result

    return result


def example_parallel_jobs2(N, Nhosts=1):

    jobserver = JobServer(hosts=[gethostname()]*Nhosts)

    for i in range(1, N+1):
        jobserver.submit_job(new_plummer_model, (i,))

    jobserver.waitall()

    result = dict()

    for job in jobserver.finished_jobs:
        result[job.args[0]] = job.result

    return result


class TestRemoteCode(amusetest.TestCase):

    def check_not_in_mpiexec(self):
        """
        The tests will fork another process, if the test run 
        is itself an mpi process, the tests may fail.

        For the hydra process manager the tests will fail.
        So skip the tests if we detect hydra
        """

        if 'HYDRA_CONTROL_FD' in os.environ or 'PMI_FD' in os.environ:
            self.skip('cannot run the socket tests under mpi process manager')

    def test1(self):
        remote = RemoteCodeInterface()

        var = 123
        remote.assign("var", var)
        var_ = remote.evaluate("var+1")
        self.assertEqual(var_, var+1)
        remote.execute("var=var*2")
        var_ = remote.evaluate("var")
        self.assertEqual(var_, var*2)

    def test2(self):
        remote = RemoteCodeInterface()

        from math import sqrt
        result = remote.func(sqrt, 64)
        self.assertEqual(result, 8)

    def test3(self):
        remote = RemoteCodeInterface()

        var = new_plummer_model(100)
        remote.assign("var", var)
        var_ = remote.evaluate("var.mass")
        self.assertEqual(var_, var.mass)

    def test4(self):
        self.check_not_in_mpiexec()
        remote = RemoteCodeInterface(channel_type="sockets")

        var = 123
        remote.assign("var", var)
        var_ = remote.evaluate("var+1")
        self.assertEqual(var_, var+1)
        remote.execute("var=var*2")
        var_ = remote.evaluate("var")
        self.assertEqual(var_, var*2)


class TestJobServer(amusetest.TestCase):

    def test1(self):

        result = example_parallel_jobs(10)
        for arg, res in result.items():
            self.assertEqual(arg, len(res))

    def test2(self):

        result = example_parallel_jobs(10, 4)
        for arg, res in result.items():
            self.assertEqual(arg, len(res))

    def test3(self):

        result = example_parallel_jobs2(10, 4)
        for arg, res in result.items():
            self.assertEqual(arg, len(res))



import os

from amuse.test.amusetest import TestWithMPI
from amuse.ext import concurrent

from amuse.rfi.core import *
from amuse.ic.plummer import new_plummer_sphere
from amuse.units import nbody_system

import numpy


class ConcurrentTestingInterface(PythonCodeInterface):

    def __init__(self, implementation_factory, **options):
        PythonCodeInterface.__init__(self, implementation_factory=implementation_factory, **options)

    @legacy_function
    def do_concurrent_run():
        function = LegacyFunctionSpecification()
        function.addParameter('output', dtype='string', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        function.id = 10
        return function


class Test1Implementation(object):

    def create_particles(self):
        self.particles = new_plummer_sphere(100)

    def do_concurrent_run(self, output):
        processes = concurrent.MPIConcurrentProcesses()
        processes.init()
        try:
            self.particles = None
            processes.on_root(self.create_particles)
            shared_particles = processes.share(self.particles)
            shared_particles.distribute()
            assert len(shared_particles) == 100, '{0} != 100'.format(len(shared_particles))
            output.value = str('success')
            result = 0
        except Exception as ex:
            output.value = str(ex)
            result = -1

        return self.reduce_result(processes, result)

    def reduce_result(self, processes, result):
        mpi_comm = processes.mpi_comm
        from mpi4py import MPI
        input = numpy.zeros(1,  dtype='int64')
        output = numpy.zeros(1,  dtype='int64')

        input[0] = result

        if mpi_comm.rank == 0:
            mpi_comm.Reduce(
                [input, MPI.INTEGER8],
                [output, MPI.INTEGER8],
                op=MPI.SUM,
                root=0
            )
        else:
            mpi_comm.Reduce(
                [input, MPI.INTEGER8],
                [output, MPI.INTEGER8],
                op=MPI.SUM,
                root=0
            )

        return output[0]


class Test3Implementation(Test1Implementation):

    def do_concurrent_run(self, output):
        processes = concurrent.MPIConcurrentProcesses()
        processes.init()
        test = TestMPIConcurrentProcesses('test1')
        try:
            self.particles = None
            processes.on_root(self.create_particles)
            shared_particles = processes.share(self.particles)
            shared_particles.distribute()
            test.assertEqual(len(shared_particles), 100)
            test.assertAlmostRelativeEquals(shared_particles.mass.sum(), 1.0 | nbody_system.mass)
            output.value = str('success')
            result = 0
        except Exception as ex:
            print(ex)
            output.value = str(ex)
            result = -1

        return self.reduce_result(processes, result)


class Test4Implementation(Test1Implementation):

    def root_potential_energy(self):
        self.potential_energy = self.particles.potential_energy(G=nbody_system.G)

    def distributed_potential_energy_on_root(self):
        test = TestMPIConcurrentProcesses('test1')
        self.distributed_potential_energy = self.shared_particles.potential_energy(G=nbody_system.G)
        test.assertAlmostRelativeEquals(self.distributed_potential_energy, self.potential_energy)

    def distributed_potential_energy_not_on_root(self):
        self.distributed_potential_energy = self.shared_particles.potential_energy(G=nbody_system.G)

    def do_concurrent_run(self, output):
        processes = concurrent.MPIConcurrentProcesses()
        processes.init()
        test = TestMPIConcurrentProcesses('test1')
        self.particles = None
        self.potential_energy = 0
        try:
            processes.on_root(self.create_particles)
            processes.on_root(self.root_potential_energy)
            self.shared_particles = processes.share(self.particles)
            self.shared_particles.distribute()
            processes.call(
                self.distributed_potential_energy_on_root,
                self.distributed_potential_energy_not_on_root,
            )
            output.value = str('success')
            result = 0
        except Exception as ex:
            print(ex)
            output.value = str(ex)
            result = -1

        return self.reduce_result(processes, result)


class TestMPIConcurrentProcesses(TestWithMPI):

    def setUp(self):
        pass

    def tearDown(self):
        os.remove("test_concurrent")

    def test1(self):
        x = ConcurrentTestingInterface(implementation_factory=Test1Implementation)
        output, error = x.do_concurrent_run()
        self.assertEqual(error, 0, msg=output)

    def test2(self):
        x = ConcurrentTestingInterface(
            implementation_factory=Test1Implementation,
            number_of_workers=4
        )
        output, error = x.do_concurrent_run()
        self.assertEqual(error, 0, msg=output)

    def test3(self):
        x = ConcurrentTestingInterface(
            implementation_factory=Test3Implementation,
            number_of_workers=4
        )
        output, error = x.do_concurrent_run()
        self.assertEqual(error, 0, msg=output)

    def test4(self):
        x = ConcurrentTestingInterface(
            implementation_factory=Test4Implementation,
            number_of_workers=4
        )
        output, error = x.do_concurrent_run()
        self.assertEqual(error, 0, msg=output)
